#ifndef __RTL_ARCH_TIME_H__
#define __RTL_ARCH_TIME_H__

/*
 * rtl_time.h
 *
 * hardware clock interface
 *
 * Written by Michael Barabanov
 * Copyright VJY Associates
 * Released under the terms of the LGPL
 *
 */

#ifdef __KERNEL__
#include <linux/time.h>
#include <linux/tasks.h>
#include <asm/ptrace.h>
#include <asm/smp.h>

#ifdef __SMP__
#define rtl_get_cpu_id() hard_smp_processor_id()
#define rtl_num_cpus() (smp_num_cpus)
#else
#define rtl_get_cpu_id() 0
#define rtl_num_cpus() 1
#endif

#define i8254_TICKS_PER_SEC 1193180UL
static const struct timespec RTL_TIME_INFINITY = {0x7fffffff, 0};
static const struct timespec RTL_TIME_ZERO = {0, 0};


#define DECLARE_CPUID(cpu_id)   unsigned int cpu_id = rtl_get_cpu_id()

#include <linux/delay.h>

extern __inline__ unsigned long muldiv(unsigned long a, unsigned long mul, unsigned long div)
{
       __asm__("mull %1 ; divl %2"
               :"=a" (a)
               :"d" (mul),
                "r" (div),
                "0" (a)
               :"dx");
       return a;
}

extern long rtl_delay_mult;

static inline void rtl_delay(long nanoseconds)
{
	__asm__("mull %0"
		:"=d" (nanoseconds)
		:"a" (nanoseconds),"0" (rtl_delay_mult)
		:"ax");
		
	__delay(nanoseconds);
}

struct clock_struct;

typedef void (* RTL_CLOCK_HANDLER)(struct pt_regs *r);

struct hw_clock;
struct hw_clock {
	int (*init) (struct hw_clock *);
	void (*uninit) (struct hw_clock *);
	int (*gettime) (struct timespec *ts);
	int (*settimer) (int mode, struct timespec *interval);
	void (*checklinuxirq) (void);
	RTL_CLOCK_HANDLER handler;
	int istimerset;
	int mode;
	int apic_cpu;
	struct timespec linux_time;
	struct timespec time;
	struct timespec resolution;
};
typedef struct hw_clock * clockid_t;

/* #define clock_getres (clock, tsptr) { *(tsptr) = clock->resolution; } */


/* modes */
#define CLOCK_MODE_UNINITIALIZED  -1
#define CLOCK_MODE_ONESHOT  0
#define CLOCK_MODE_PERIODIC 1

#define rtl_clock_settimer(clock, mode, tsptr) ((clock)->settimer( mode, tsptr))
#define rtl_clock_getmode(clock) ((clock)->mode)


/* maximum interval between interrupts in nanoseconds */
#define RTL_MAX_INTERVAL (NSECS_PER_SEC/(HZ * 2))

/* scheduler interface */
struct hw_clock * rtl_best_clock(unsigned int ,RTL_CLOCK_HANDLER );
extern int rtl_clock_sethandler (struct hw_clock *, RTL_CLOCK_HANDLER);
extern int rtl_gettime (struct timespec *);
extern int rtl_clock_unsethandler (struct hw_clock *);
extern int rtl_clock_ishandlerset (struct hw_clock *);
#define rtl_clock_checklinuxirq(clock) {clock->checklinuxirq();}
/* end of scheduler interface */


/* for rtl_core */
extern void rtl_hard_enable_irq(int ix);
extern void rtl_hard_disable_irq(int ix);

#ifdef __SMP__
extern int request_ipi (unsigned int (*f)(struct pt_regs *r), int cpu);
extern int free_ipi (int cpu);
#endif



extern struct hw_clock * rtl_clock_init(RTL_CLOCK_HANDLER );
extern void rtl_clock_uninit (struct hw_clock *);

extern int I8253_channel2_free(void);

#endif

#define NSECS_PER_SEC 1000000000
#define timespec_add(t1, t2) { \
	(t1)->tv_nsec += (t2)->tv_nsec;  \
	(t1)->tv_sec += (t2)->tv_sec; \
	if ((t1)->tv_nsec >= NSECS_PER_SEC) { \
		(t1)->tv_nsec -= NSECS_PER_SEC; \
		(t1)->tv_sec++; \
	} else if ((t1)->tv_nsec < 0) { \
		(t1)->tv_nsec += NSECS_PER_SEC; \
		(t1)->tv_sec--; \
	} \
}

#define timespec_sub(t1, t2) { \
	(t1)->tv_nsec -= (t2)->tv_nsec;  \
	(t1)->tv_sec -= (t2)->tv_sec; \
	if ((t1)->tv_nsec < 0) { \
		(t1)->tv_nsec += NSECS_PER_SEC; \
		(t1)->tv_sec--; \
	} else if ((t1)->tv_nsec >= NSECS_PER_SEC) { \
		(t1)->tv_nsec -= NSECS_PER_SEC; \
		(t1)->tv_sec++; \
	} \
}

#define timespec_nz(t) ((t)->tv_sec != 0 || (t)->tv_nsec != 0)

#define timespec_lt(t1, t2) ((t1)->tv_sec < (t2)->tv_sec || ((t1)->tv_sec == (t2)->tv_sec && (t1)->tv_nsec < (t2)->tv_nsec))

#define timespec_gt(t1, t2) (timespec_lt(t2, t1))

#define timespec_ge(t1, t2) (!timespec_lt(t1, t2))

#define timespec_le(t1, t2) (!timespec_gt(t1, t2))

#define timespec_eq(t1, t2) ((t1)->tv_sec == (t2)->tv_sec && (t1)->tv_nsec == (t2)->tv_nsec)

static inline long long timespec_to_ns (struct timespec *ts)
{
	long long t;

	__asm__("imull %%edx\n"
		"add %%ebx, %%eax\n\t"
		"adc $0, %%edx\n\t"
		:"=A" (t)
		: "a" (ts->tv_sec), "d" (NSECS_PER_SEC), "b" (ts->tv_nsec)
		: "ax", "dx");
	return t;
}

static inline struct timespec timespec_from_ns (long long t)
{
	struct timespec ts;
	__asm__("idivl %%ecx\n\t"
		:"=a" (ts.tv_sec), "=d" (ts.tv_nsec)
		: "A" (t), "c" (NSECS_PER_SEC)
		:"cx");
	if (ts.tv_nsec < 0) {
		ts.tv_nsec += NSECS_PER_SEC;
		ts.tv_sec --;
	}
	return ts;
}


#endif

